/*
 * Copyright 2010 Trustees of the University of Pennsylvania Licensed under the
 * Educational Community License, Version 2.0 (the "License"); you may
 * not use this file except in compliance with the License. You may
 * obtain a copy of the License at
 *
 * http://www.osedu.org/licenses/ECL-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */
package metridoc.plugins.sql

import groovy.sql.Sql
import javax.sql.DataSource
import metridoc.utils.Assert
import metridoc.utils.PropertyUtils
import org.apache.commons.lang.StringEscapeUtils
import org.apache.commons.lang.math.NumberUtils
import org.slf4j.LoggerFactory

/**
 * Created by IntelliJ IDEA.
 * User: tbarker
 * Date: 7/18/11
 * Time: 3:38 PM
 */
class SqlPlus extends Sql {

    static final PHASE_NAMES = "phaseName"
    static final log = LoggerFactory.getLogger(SqlPlus.class)
    static final Set IGNORED_KEYS = [PHASE_NAMES, "order"]

    SqlPlus(DataSource dataSource) {
        super(dataSource)
    }

    void runBatchFile(LinkedHashMap args) {
        Assert.notNull(args.fileName, "SqlPlus requires a fileName to runBatchFile")
        ConfigObject file = new PropertyUtils().getConfig(args.fileName)
        def phases = getPhases(file)
        phases.each {order, phase ->
            long beginPhaseTime = new Date().getTime()
            def phaseName = phase.phaseName
            log.info "starting phase ${phaseName}"
            if (!exclude(args.exclude, phaseName)) {
                phase.each {sqlName, value ->
                    def fullSqlName = "${phaseName}.${sqlName}"
                    if (!exclude(args.exclude, fullSqlName)) {

                        if (!IGNORED_KEYS.contains(sqlName)) {
                            log.info("running ${fullSqlName}")
                            long startTime = new Date().getTime()
                            int updateCount = executeUpdate(value.sql)
                            long endTime = new Date().getTime()
                            long totalTime = endTime - startTime
                            log.info("finished running ${fullSqlName} with ${updateCount} updates, took ${totalTime} milliseconds")
                        }
                    } else {
                        log.info("skipping sql ${fullSqlName}")
                    }

                }
            } else {
                log.info("skipping phase ${phaseName}")
            }

            long endPhaseTime = new Date().getTime()
            long totalPhaseTime = endPhaseTime - beginPhaseTime
            log.info "finished ${phaseName}, took ${totalPhaseTime} milliseconds"
        }
    }

    private static boolean exclude(exclude, String value) {
        if (exclude) {
            return exclude.contains(value)
        }

        return false
    }

    private static SortedMap getPhases(ConfigObject configObject) {
        def result = new TreeMap()
        int defaultOrder = 1000
        ConfigObject sqlPhases = configObject.sql
        sqlPhases.each {key, value ->
            try {
                value[PHASE_NAMES] = key
                value.order = value.containsKey("order") ? value.order : defaultOrder
                result.put(Double.valueOf(value.order.toString()), value)
            } catch (Exception ex) {
                throw new RuntimeException("there was an error loading phase ${key}", ex)
            }
        }

        return result
    }


    int[] insertBatch(String table, batch) {

        withBatch {statement ->
            if (batch instanceof List) {
                batch.each {record ->

                    if (record == null) {
                        throw new IllegalArgumentException("a record must be a none null Map to use batch inserting")
                    }

                    if (!(record instanceof Map)) {
                        throw new IllegalArgumentException("record ${record} must be of type Map to use batch inserting")
                    }

                    def insert = getInsertStatement(table, record)
                    log.trace("adding {} to batch inserts", insert)
                    statement.addBatch(insert)
                }
            } else {
                def insert = getInsertStatement(table, batch)
                if (batch == null) {
                    throw new IllegalArgumentException("a record must be a none null Map to use batch inserting")
                }

                if (!(batch instanceof Map)) {
                    throw new IllegalArgumentException("record ${record} must be of type Map to use batch inserting")
                }

                log.trace("adding {} to batch inserts", insert)
                statement.addBatch(insert)
            }
        }
    }

    private static String getInsertStatement(String table, Map values) {

        StringBuffer insert = new StringBuffer("insert into ")
        StringBuffer valuesToInsert = new StringBuffer("values (")
        insert.append(table)
        insert.append(" (")
        values.each {key, value ->
            insert.append(key)
            insert.append(", ")
            String valueText = String.valueOf(value)

            if (NumberUtils.isNumber(valueText)) {
                valuesToInsert.append(value)
            } else {
                valuesToInsert.append("'")
                valuesToInsert.append(StringEscapeUtils.escapeSql(value))
                valuesToInsert.append("'")
            }

            valuesToInsert.append(", ")
        }
        insert.delete(insert.length() - 2, insert.length())
        insert.append(") ")
        valuesToInsert.delete(valuesToInsert.length() - 2, valuesToInsert.length())
        valuesToInsert.append(")")

        return insert.append(valuesToInsert).toString()
    }


}

class BatchInWrongFormatException extends RuntimeException {
    Class wrongType

}